# animatedviz.py

import matplotlib
from matplotlib import cm
import matplotlib.patches
import matplotlib.transforms
from numpy import pi
import numpy as np
import torch

import matplotlib.animation as animation

import matplotlib.pyplot as plt

from intersim.viz.utils import batched_rotate_around_center, draw_map_without_lanelet

import os

opj = os.path.join

class AnimatedViz:
    '''
    Animated visualization of a state sequence
    '''

    def __init__(self, ax, osm, states, lengths, widths, graphs=None):
        '''
        Args:
            ax (plt.Axis): matplotlib Axis
            osm (str): osm file name
            states (torch.tensor): (T, nv*5) states
            lengths (torch.tensor): (nv,) car lengths
            widths (torch.tensor): (nv,) car widths
            graphs (list of list of tuples): graphs[i][j] is the jth edge tuple in the ith frame
        ''' 

        self._T = states.shape[0]
        self._nv = states.shape[1] // 5

        states = states.reshape(self._T,self._nv,5)

        self._ax = ax
        self._osm = osm
        self._x = states[...,0].detach().numpy()
        self._y = states[...,1].detach().numpy()
        self._psi = states[...,3].detach().numpy()

        self._lengths = lengths.detach().numpy()
        self._widths = widths.detach().numpy()
        
        self._graphs = graphs
        
    @property
    def lines(self):
        return self._carrects + self._edges + [self._text]

    def initfun(self):
        ax = self._ax 

        draw_map_without_lanelet(self._osm, ax, 0.,0.)
        ax.axes.xaxis.set_visible(False)
        ax.axes.yaxis.set_visible(False)

        # init car patches
        carrects = []
        cmap = cm.get_cmap('jet')
        car_colors = cmap(np.linspace(0,1,num=self._nv))
        np.random.seed(0)
        np.random.shuffle(car_colors)
        for i in range(self._nv):
            rectpts = np.array([(-1.,-1.), (1.,-1), (1.,1.), (-1.,1.)])
            rect = matplotlib.patches.Polygon(rectpts, closed=True, color=car_colors[i], zorder=9, ec='k')
            ax.add_patch(rect)
            carrects.append(rect)
        self._carrects = carrects
        self._edges = []
        self._text = ax.text(0.05, 0.9, '', transform=ax.transAxes)

        return self.lines

    def animate(self, i):
        '''
        Args:
            i (int): animation timestep
        '''

        x, y, lengths, widths = self._x, self._y, self._lengths, self._widths
        ax = self._ax
        psi = self._psi

        nni = ~np.isnan(x[i])
        x = x[i, nni]
        y = y[i, nni]
        lengths = lengths[nni]
        widths = widths[nni]
        psi = psi[i, nni]

        self._text.set_text('i=%d' % i)
            
        T = self._T

        i = min(T-1, i)

        lowleft = np.stack([x - lengths / 2., y - widths / 2.], axis=-1)
        lowright = np.stack([x + lengths / 2., y - widths / 2.], axis=-1)
        upright = np.stack([x + lengths / 2., y + widths / 2.], axis=-1)
        upleft = np.stack([x - lengths / 2., y + widths / 2.], axis=-1)

        rotcorners = batched_rotate_around_center(np.stack([lowleft, lowright, upright, upleft],axis=1), 
                np.stack([x, y], axis=-1), yaw=psi)

        all_corners = np.stack([np.array([(-1.,-1.), (1.,-1), (1.,1.), (-1.,1.)])]*self._nv)
        all_corners[nni] = rotcorners

        for corners, carrect in zip(all_corners, self._carrects):
            carrect.set_xy(corners)
        
        edges = []
        if self._graphs is not None:
            for edge in self._edges:
                edge.remove()
            
            graph = self._graphs[i]
            for e in graph:
                stidx, enidx = e
                
                ars = '<|-|>' if (enidx, stidx) in graph else '-|>'
                arrow = matplotlib.patches.FancyArrowPatch(posA = (self._x[i,stidx], self._y[i,stidx]),
                    posB = (self._x[i,enidx], self._y[i,enidx]),
                    arrowstyle=ars, mutation_scale=15, color='w', zorder=9.5, ec='k',)

                ax.add_patch(arrow)
                edges.append(arrow)
        self._edges = edges
        return self.lines


# box.py

# Adapted from https://github.com/openai/gym/blob/master/gym/spaces/box.py

import numpy as np
import torch
from torch.distributions.exponential import Exponential

class Box:
    """
    A (possibly unbounded) box in R^n. Specifically, a Box represents the
    Cartesian product of n closed intervals. Each interval has the form of one
    of [a, b], (-oo, b], [a, oo), or (-oo, oo).

    There are two common use cases:

    * Identical bound for each dimension::
        >>> Box(low=-1.0, high=2.0, shape=(3, 4), dtype=np.float32)
        Box(3, 4)

    * Independent bound for each dimension::
        >>> Box(low=np.array([-1.0, -2.0]), high=np.array([2.0, 4.0]), dtype=np.float32)
        Box(2,)

    """
    def __init__(self, low, high, shape=None, dtype=torch.get_default_dtype()):
        assert dtype is not None, 'dtype must be explicitly provided. '
        self.dtype = dtype

        # determine shape if it isn't provided directly
        if shape is not None:
            shape = tuple(shape)
            assert np.isscalar(low) or low.shape == shape, "low.shape doesn't match provided shape"
            assert np.isscalar(high) or high.shape == shape, "high.shape doesn't match provided shape"
        elif not np.isscalar(low):
            shape = low.shape
            assert np.isscalar(high) or high.shape == shape, "high.shape doesn't match low.shape"
        elif not np.isscalar(high):
            shape = high.shape
            assert np.isscalar(low) or low.shape == shape, "low.shape doesn't match high.shape"
        else:
            raise ValueError("shape must be provided or inferred from the shapes of low or high")

        if np.isscalar(low):
            low = torch.full(shape, low, dtype=dtype)

        if np.isscalar(high):
            high = torch.full(shape, high, dtype=dtype)

        self.shape = shape
        self.low = low
        self.high = high
        if not isinstance(low, torch.Tensor):
            self.low = torch.tensor(low)
        if not isinstance(high, torch.Tensor):
            self.high = torch.tensor(high)

        self.low = self.low.type(self.dtype)
        self.high = self.high.type(self.dtype)

        # Boolean arrays which indicate the interval type for each coordinate
        self.bounded_below = -np.inf < self.low
        self.bounded_above = np.inf > self.high

        

    def is_bounded(self, manner="both"):
        below = torch.all(self.bounded_below)
        above = torch.all(self.bounded_above)
        if manner == "both":
            return below and above
        elif manner == "below":
            return below
        elif manner == "above":
            return above
        else:
            raise ValueError("manner is not in {'below', 'above', 'both'}")

    def sample(self):
        """
        Generates a single random sample inside of the Box.

        In creating a sample of the box, each coordinate is sampled according to
        the form of the interval:

        * [a, b] : uniform distribution
        * [a, oo) : shifted exponential distribution
        * (-oo, b] : shifted negative exponential distribution
        * (-oo, oo) : normal distribution
        """

        high = self.high 
        sample = torch.empty(self.shape)

        # Masking arrays which classify the coordinates according to interval
        # type
        unbounded   = ~self.bounded_below & ~self.bounded_above
        upp_bounded = ~self.bounded_below &  self.bounded_above
        low_bounded =  self.bounded_below & ~self.bounded_above
        bounded     =  self.bounded_below &  self.bounded_above


        # Vectorized sampling by interval type
        sample[unbounded] = torch.randn(self.shape)[unbounded]

        sample[low_bounded] = Exponential(1).sample(
            self.shape)[low_bounded] + self.low[low_bounded]

        sample[upp_bounded] = -Exponential(1).sample(
            self.shape)[upp_bounded] + self.high[upp_bounded]

        sample[bounded] = torch.rand(self.shape)[bounded]*(
                        self.high[bounded]-self.low[bounded]) + self.low[bounded]


        return sample.type(self.dtype)

    def contains(self, x):
        if isinstance(x, list):
            x = torch.tensor(x)  # Promote list to array for contains check
        return x.shape == self.shape and torch.all(x >= self.low) and torch.all(x <= self.high)

    def __repr__(self):
        return "Box({}, {}, {}, {})".format(self.low.min(), self.high.max(), self.shape, self.dtype)

    def __eq__(self, other):
        return isinstance(other, Box) and (self.shape == other.shape) and torch.allclose(self.low, other.low) and torch.allclose(self.high, other.high)


#closestobstacle.py

import torch
from torch import nn
from intersim.graph import InteractionGraph
import numpy as np

class ClosestObstacleGraph(InteractionGraph):
    """
    Observe the closest vehicle within half-angle degrees.
    """
    
    def __init__(self, half_angle = 20, neighbor_dict = {}):
        """
        Args:
            half_angle (float): half_angle for observation
            neighbor_dict (dict of vertices to edges): dictionary of node neighbors
        """
        InteractionGraph.__init__(self, neighbor_dict)

        self._half_angle = half_angle

    def update_graph(self, state):
        """
        Update the neighbor dict given a new state.
        
        Args:
            state (torch.tensor): (nv*5,) vehicle states
        """
        ndim = state.ndim
        assert ndim==1, "Improper state size to graph updater"
        
        state = state.reshape(-1, 5)
        nv = state.shape[-2]
        
        x = state[...,0:1]
        y = state[...,1:2]
        v = state[...,2:3]
        psi = state[...,3:4]
        psidot = state[...,4:5]

        # compute relative distances, relative velocities
        dx = x.unsqueeze(-2) - x.unsqueeze(-3) #(*, nv, nv, 1) where dx[*,i,j] is x[*,i] - x[*,j]
        dy = y.unsqueeze(-2) - y.unsqueeze(-3)
        dr = (dx ** 2 + dy ** 2).sqrt()
        delpsi = self.to_circle(torch.atan2(dy,dx) - psi.unsqueeze(-3))
        ndist = dr * delpsi.cos()
        ndisth = dr * delpsi.sin()
        
        # now select the vehicles within distance and half-angle requirements
        ndist_ = torch.where((ndist > 0) & (delpsi.abs() < self._half_angle*np.pi/180), ndist, np.inf)

        cndist, inds = ndist_.min(dim=-3)
        
        # generate and store neighbor graph
        neighbor_dict = {}
        for i in range(nv): 
            if cndist[i,0] < np.inf:
                neighbor_dict[i] = [inds[i,0]]
            
        self._neighbor_dict = neighbor_dict
        
    def to_circle(self, x):
        """
        Casts x (in rad) to [-pi, pi)
        
        Args:
            x (torch.tensor): (*) input angles (radians)
            
        Returns:
            y (torch.tensor): (*) x cast to [-pi, pi)
        """
        y = torch.remainder(x + np.pi, 2*np.pi) - np.pi
        return y


#
# collisions.py
#

import torch
import numpy as np
from shapely.geometry import Polygon


def states_to_polygons(x, lengths, widths):
    """
    Create a set of polygons representing each active vehicle.
    Args:
        x (torch.tensor): (nv, 5) vehicle states
        legnths (torch.tensor): (nv,) vehicle lengths
        widths (torch.tensor): (nv,) vehicle lengths
    Returns
        vp (list of Polygon): (nv_active,) vehicle Polygons
        inds (torch.tensor): (n,) indices of active vehicles
    """

    nni = ~torch.isnan(x[:,0])
    x = x[nni]
    vp = []
    for  x_, li, wi in zip(x, lengths[nni], widths[nni]):
        px, py, v, psi, psidot = x_

        pxy = torch.stack([px, py])
        lon = torch.stack([psi.cos(), psi.sin()])
        lat = torch.stack([-psi.sin(), psi.cos()])
        
        ul = pxy + li/2. * lon + wi/2. * lat
        ur = pxy + li/2. * lon - wi/2. * lat
        ll = pxy - li/2. * lon + wi/2. * lat
        lr = pxy - li/2. * lon - wi/2. * lat
        
        corners = torch.stack([ll, lr, ur, ul]).detach().numpy()
        p = Polygon([*corners])
        vp.append(p)
    inds = torch.where(nni)[0]
    return vp, inds



def count_collisions(x, lengths, widths):
    """
    Count collisions in a trajectory.
    Args:
        x (torch.tensor): (T,nv*5) vehicle state
        legnths (torch.tensor): (nv,) vehicle lengths
        widths (torch.tensor): (nv,) vehicle lengths
    Returns
        ncols (int): number of collisions
    """

    T,_ = x.shape
    x = x.reshape(T, -1, 5)
    nv0 = x.shape[1]
    cols = torch.zeros((nv0, nv0))
    for t in range(T):
        polys_t, nni_t = states_to_polygons(x[t], lengths, widths)
        nv = len(polys_t)
        for i in range(1,nv):
            for j in range(i):
                if polys_t[i].intersects(polys_t[j]):
                    cols[nni_t[i],nni_t[j]] += 1
                    
    cols = cols + cols.T # make symmetric 
    ncols = np.count_nonzero(cols) / 2 # divide by 2 since symmetric
    return ncols



# graph.py

class InteractionGraph:
    
    def __init__(self, neighbor_dict={}):
        self._neighbor_dict = neighbor_dict
    
    def update_graph(self, x):
        """
        Update the neighbor dict given a new state.
        
        Args:
            x (torch.tensor): (nv*5,) vehicle states
        """
        raise NotImplementedError
    
    @property
    def neighbor_dict(self):
        return self._neighbor_dict.copy()
    
    @property
    def nodes(self):
        return sorted(list(self._neighbor_dict.keys()))
    
    @property
    def edges(self):
        return [(i,j) for i in self.nodes for j in self._neighbor_dict[i]]
    
    @property
    def reverse_edges(self):
        return [(j,i) for i in self.nodes for j in self._neighbor_dict[i]]
    
    @property
    def reverse_neighbor_dict(self):
        rdict = {node:[] for node in self.nodes}
        for edge in self.reverse_edges:
            rdict[edge[0]].append(edge[1]) 
        return rdict
    
    @property
    def strongly_connected_components(self):
        """
        Compute the strongly connected components of the graph.
        Args:
        Returns:
            sccs (list of tuples): first element is vertices in each scc, second is outgoing edges
        """
        start = {n:-1 for n in self.nodes} # forward dfs start times
        fin = {n:-1 for n in self.nodes} # forward dfs end times
        self._visit_time = 0 
        for node in self.nodes: 
            if start[node] == -1: # if unvisited, recurse on subtree
                self.DFS(node, self._neighbor_dict, start, fin, [])
                
        # compute reverse neighbor dict
        rdict = self.reverse_neighbor_dict
        
        # compute node ordering based on reverse finish time
        sorted_items = sorted(fin.items(), key=lambda item:-item[1])
        reverse_ordering = [item[0] for item in sorted_items]
        
        start = {n:-1 for n in self.nodes} # backward dfs start times
        fin = {n:-1 for n in self.nodes} # backward dfs end times
        self._visit_time = 0 
        scc_list = []
        for node in reverse_ordering:
            if start[node] == -1: # new scc
                subtree = [node]
                self.DFS(node, rdict, start, fin, subtree)
                scc_list.append(sorted(subtree))
        
        # find outgoing edges
        sccs = []
        for scc in scc_list:
            outgoing_edges = []
            for node in scc:
                for neighbor_node in self._neighbor_dict[node]:
                    if neighbor_node not in scc and neighbor_node not in outgoing_edges:
                        outgoing_edges.append(neighbor_node)
            sccs.append((scc,sorted(outgoing_edges)))       
        return sccs
            
        
    def DFS(self, node, neighbor_dict, start, fin, tree):
        """
        Perform a depth-first search of a subtree.
        Args:
            node (int): current root node of search
            neighbor_dict (dict): neighbor dictionary to use for traversal
            start (dict): dict of search start times
            fin (dict): dict of search end times
            tree (list): list of nodes in current tree
        Returns:
        """    

        self._visit_time += 1
        start[node] = self._visit_time
        for neighbor in neighbor_dict[node]:
            if start[neighbor] == -1: # recurse if unexplored
                tree.append(neighbor) # add neighbor to current subtree
                self.DFS(neighbor, neighbor_dict, start, fin, tree)
        fin[node] = self._visit_time



# vehicletraj.py

import numpy as np
from numpy.polynomial.polynomial import Polynomial as P
from numpy.polynomial.polynomial import polyder
import torch

class StackedVehicleTraj:

	def __init__(self, lengths, widths, t0, s, v, xpoly, ypoly, dt=0.1):
		"""
		Stacked vehicle trajectories as polynomials.
		Args:
			lengths (torch.tensor): (nv,) vehicle lengths
			widths (torch.tensor): (nv,) vehicle widths
			t0 (torch.tensor): (nv,) start times
			s (list of torch.tensor): list of nv path positions
			v (list of torch.tensor): list of nv path velocities
			xpoly (torch.tensor): (nv, K) polynomial coefficients of x(s)
			ypoly (torch.tensor): (nv, K) polynomial coefficients of y(s)
			dt (float): timestep
		"""

		self._lengths = lengths
		self._widths = widths

		self._nv = len(lengths)
		self._t0 = t0
		self._s = s
		self._v = v
		self._smax = torch.tensor([s_[-1] for s_ in s])
		self._v0 = torch.tensor([v_[0] for v_ in v])

		self._xpoly = xpoly.detach()
		self._ypoly = ypoly.detach()

		# compute dxpoly, dypoly, ddxpoly, ddypoly
		def difpoly(poly):
			poly = poly.detach().numpy()
			dpoly = torch.zeros(*poly.shape).type(xpoly.type())
			ddpoly = torch.zeros(*poly.shape).type(xpoly.type())
			for i, p in enumerate(poly):

				dpolser = polyder(p).copy()
				dpolser = torch.tensor(dpolser)
				dpoly[i,:-1] = dpolser

				ddpolser = polyder(p, 2).copy()
				ddpolser = torch.tensor(ddpolser)
				ddpoly[i,:-2] = ddpolser

			return dpoly, ddpoly

		self._dxpoly, self._ddxpoly = difpoly(xpoly)
		self._dypoly, self._ddypoly = difpoly(ypoly)

		self._dt = dt

		# compute min, max T
		minT = float(t0.min())
		minTind = int(minT/dt)
		maxTind = max([ len(s_) + int(t0_/dt) for (s_,t0_) in zip(s,t0) ])
		maxT = maxTind * dt
		Tind = maxTind - minTind
		T = maxT - minT

		self._minT = minT
		self._maxT = maxT
		self._minTind = minTind
		self._maxTind = maxTind
		self._Tind = Tind
		self._T = T

		# initial states
		state0 = torch.zeros(self._nv, 2) * np.nan
		for i in range(self._nv): 
			if self._t0[i] == minT:
				state0[i,0] = s[i][0]
				state0[i,1] = v[i][0]

		self._state0 = state0

	@property
	def state0(self):
		return self._state0

	@property
	def polydeg(self):
		return self._xpoly.shape[-1]-1
	
	@property
	def xpoly(self):
		return self._xpoly
			
	@property
	def ypoly(self):
		return self._ypoly
	
	@property
	def dxpoly(self):
		return self._dxpoly
	
	@property
	def dypoly(self):
		return self._dypoly
	
	@property
	def ddxpoly(self):
		return self._ddxpoly
	
	@property
	def ddypoly(self):
		return self._ddypoly
		
	@property
	def T(self):
		return self._T
	
	@property
	def Tind(self):
		return self._Tind

	@property
	def minT(self):
		return self._minT

	@property
	def minTind(self):
		return self._minTind
	
	@property
	def s(self):
		return self._s

	@property
	def v(self):
		return self._v
	
	@property
	def v0(self):
		return self._v0

	@property
	def smax(self):
		return self._smax

	@property
	def t0(self):
		return self._t0

	@property
	def nv(self):
		return self._nv

	@property
	def dt(self):
		return self._dt

	@property
	def lengths(self):
		return self._lengths
	
	@property
	def widths(self):
		return self._widths
	
	
	
	


# simulator.py

import torch

from intersim.datautils import ssdot_to_simstates
from intersim import Box
from intersim import StackedVehicleTraj
import numpy as np

class RoundaboutSimulator():

    def __init__(self, svt, max_acc=1.0):
        """
        Roundabout simulator.
        Args:
            svt (StackedVehicleTraj): vehicle trajectory to reference
            max_acc (float): maximum acceleration/deceleration
        Notes:
            action_space is [acc] ^ nvehicles
            state_space is [x,y,v,psi,psidot] ^ nvehicles
        """

        self._nv = svt.nv
        self._dt = svt.dt
        self._svt = svt

        super(RoundaboutSimulator, self).__init__()
        self.action_space = Box(low=-max_acc, high=max_acc, shape=(self._nv,1))

        low = torch.tensor([[-np.inf,-np.inf,0.,-np.pi, -np.inf]]).expand(self._nv,5)
        high = torch.tensor([[np.inf,np.inf,np.inf,np.pi, np.inf]]).expand(self._nv,5)
        self.state_space = Box(low=low, high=high, shape=(self._nv,5))

        self._state = torch.zeros(self._nv, 2) * np.nan
        self._t = svt.minT

        self._exceeded = None
    
    @property
    def state(self):
        """
        Return [s, v]
        Returns:
            state (torch.tensor): (nv, 2) state in terms of s, sdot
        """
        return self._state
    
    @property
    def path_coefs(self):
        """
        Return coefficients for computing x(s) and y(s)
        Returns:
            xpoly (torch.tensor): (nv, deg) x polynomial path coefficients
            ypoly (torch.tensor): (nv, deg) y polynomial path coefficients
        """
        svt = self._svt
        return svt.xpoly, svt.ypoly
    
    @property
    def smax(self):
        """
        Return maximum path lengths
        Returns:
            smax (torch.tensor): (nv,) maximum path lengths
        """
        return self._svt.smax
    
    @property
    def projected_state(self):
        """
        Projects [s, v] to [x,y,v,psi,psidot]
        Returns:
            projstate (torch.tensor): (nv, 5) projected state
        """
        svt = self._svt
        projstate = ssdot_to_simstates(self._state[:,0].unsqueeze(0),self._state[:,1].unsqueeze(0),
                svt.xpoly, svt.dxpoly, svt.ddxpoly,
                                svt.ypoly, svt.dypoly, svt.ddypoly)
        return projstate[0].reshape(self._nv, 5)
    

    def step(self, action):
        """
        Step simulation forward 1 time-step.
        Args:
            action (torch.tensor): (nvehicles,1) accelerations
        Returns:
            next_state (torch.tensor): (nvehicles*5,) next state
            info (dict): info
        """

        self._t += self._dt

        # euler step the state
        nextv = self._state[:,1:2] + action * self._dt
        nextv = nextv.clamp(0., np.inf) # not differentiable!
        self._state[:,0:1] = self._state[:,0:1] + 0.5*self._dt*(nextv + self._state[:,1:2])
        self._state[:,1:2] = nextv

        nni = ~torch.isnan(self._state[:,0])      

        # see which states exceeded their maxes
        self._exceeded = (self._state[:,0] > self._svt.smax) | self._exceeded
        self._state[self._exceeded] = np.nan

        # see which new vehicles should spawn
        should_spawn = (self._svt.t0 < self._t) & ~self._exceeded & torch.isnan(self._state[:,0])

        # for now, just spawn them
        spawned = should_spawn
        self._state[spawned,0] = 0.
        self._state[spawned,1] = self._svt.v0[spawned]


        return self.projected_state.reshape(-1), {
                        'exceeded':np.argwhere(self._exceeded.detach().numpy()).flatten(),
                         'should_spawn': np.argwhere(should_spawn.detach().numpy()).flatten(),
                         'spawned': np.argwhere(spawned.detach().numpy()).flatten(),
                         'raw_state': self._state.clone()}

    def reset(self):
        """
        Reset simulation.
        Returns:
            next_state (torch.tensor): (nvehicles*5,) next state
            info (dict): info
        """

        self._state = self._svt.state0
        self._t = self._svt.minT

        self._exceeded = torch.tensor([False] * self._nv)

        return self.projected_state.reshape(-1), {'raw_state': self._state.clone()}

    def render(self):
        raise NotImplementedError


# policy

import torch

class Policy:

	def compute_action(self, x):
		"""
		Abstract policy computing accelerations for each vehicle.
		Args:
			x (torch.tensor): (*, nv*5) vehicle states
		Returns:
			a (torch.tensor): (*,nv,1) actions
		"""
		raise NotImplementedError

	def __call__(self, x):
		"""
		Abstract policy computing accelerations for each vehicle.
		Args:
			x (torch.tensor): (*, nv*5) vehicle states
		Returns:
			a (torch.tensor): (*,nv,1) actions
		"""
		nv = x.shape[-1] // 5
		if x.ndim == 1:
			return self.compute_action(x.unsqueeze(0)).reshape(nv,1)

		else:
			return self.compute_action(x)


#
# File: setup.py
#
from setuptools import setup, find_packages

with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

setup(
    name="InteractionSimulator",
    version="0.0.2",
    packages=find_packages(),
    description="Simulator environment for INTERACTION driving dataset",
    long_description=long_description,
    long_description_content_type="text/markdown",
    author="Kunal Menda, Arec Jamgochian",
    author_email="jamgochian95@gmail.com",
    url="https://github.com/sisl/InteractionSimulator",
    install_requires=[
        "click",
        "joblib",
        "pandas",
        "matplotlib",
        "numpy",
        "scipy",
        "torch",
        "termcolor",
        "python_dateutil",
        "gym",
        "future",
        "sklearn",
        "pyproj",
        "pytest",
    ],    
    python_requires=">=3.6",
    test_requires=["pytest"],
    zip_safe=True,
    license="MIT",
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
        "Development Status :: 4 - Beta",
    ],)


# datautils.py

import pandas as pd
import numpy as np
from sklearn.linear_model import Lasso
import torch
from intersim.vehicletraj import StackedVehicleTraj

torch.set_default_dtype(torch.float64)

def powerseries(x, deg):

    return torch.stack([x**i for i in range(deg+1)],dim=-1)

def df_to_stackedvehicletraj(df):
    """
    Convert a vehicle_tracks dataframe to a StackedVehicleTraj
    Args:
        df (pd.DataFrame): vehicle_tracks dataframe
    Returns:
        svt (StackedVehicleTraj): stacked vehicle traj
    """

    lengths = []
    widths = []

    t0list = []
    slist = []
    xlist = []
    ylist = []
    vlist = []

    for tid in df.track_id.unique():

        df_ = df.loc[df.track_id == tid]

        lengths.append(df_.length.values[0])
        widths.append(df_.width.values[0])

        assert len(df_)-1 == (df_.frame_id.max() - df_.frame_id.min()), 'not contigous?'

        t0 = df_.timestamp_ms.min()/1000
        x = df_.x.values[:-1]
        y = df_.y.values[:-1]
        v = (df_.vx.values[:-1] ** 2 + df_.vy.values[:-1] ** 2) ** (1./2)

        dx = df_.x.diff().values[1:]
        dy = df_.y.diff().values[1:]
        ds = (dx ** 2 + dy ** 2) ** (1./2)

        s = np.cumsum(ds)

        # to tensors
        x = torch.tensor(x)
        y = torch.tensor(y)
        v = torch.tensor(v)
        s = torch.tensor(s)

        t0list.append(t0)
        slist.append(s)
        xlist.append(x)
        ylist.append(y)
        vlist.append(v)

    t0 = torch.tensor(t0list)
    lengths = torch.tensor(lengths)
    widths = torch.tensor(widths)

    xpoly, ypoly = polyfit_sxy(slist,xlist,ylist)

    dt = df_.timestamp_ms.diff().mean()/1000

    return StackedVehicleTraj(lengths, widths, t0, slist, vlist, xpoly, ypoly, dt=dt)


def polyfit_sxy(s, x, y, deg=20):
    """
    Map vehicle trajectories to xpoly, ypoly using np.polyfit
    Args:
        s (list of torch.tensor): list of nv path positions
        x (list of torch.tensor): list of nv path xs
        y (list of torch.tensor): list of nv path ys
    Returns:
        xpoly (torch.tensor): (nv, deg+1) coeffs of x(s)
        ypoly (torch.tensor): (nv, deg+1) coeffs of y(s)
    """
    nv = len(s)
    xpoly = torch.zeros(nv, deg + 1).type(torch.get_default_dtype())
    ypoly = torch.zeros(nv, deg + 1).type(torch.get_default_dtype())
    for i, (s_,x_,y_) in enumerate(zip(s,x,y)):
        s_ = s_.detach().numpy()
        x_ = x_.detach().numpy()
        y_ = y_.detach().numpy()
        pfx = np.polyfit(s_,x_,deg)[::-1].copy()
        pfy = np.polyfit(s_,y_,deg)[::-1].copy()
        xpoly[i,:] = torch.tensor(pfx)
        ypoly[i,:] = torch.tensor(pfy)

    return xpoly, ypoly

def ssdot_to_simstates(s, sdot, 
                      xpoly, dxpoly, ddxpoly, 
                      ypoly, dypoly, ddypoly):
    """
    Maps s, sdot to [x,y,v,psi,psidot] using poly coefficients.
    Args:
        s (torch.tensor): (T,nv) arc lengths
        sdot (torch.tensor): (T,nv) velocities
    Returns:
        simstates (torch.tensor): (T, nv * 5) states
    """

    nv = xpoly.shape[0]
    deg = xpoly.shape[-1] - 1
    T = s.shape[0]

    simstates = torch.ones(T, nv, 5) * np.nan
    expand_sims = powerseries(s, deg)

    x = (xpoly.unsqueeze(0)*expand_sims).sum(dim=-1)
    dxds = (dxpoly.unsqueeze(0)*expand_sims).sum(dim=-1)
    ddxds = (ddxpoly.unsqueeze(0)*expand_sims).sum(dim=-1)
    y = (ypoly.unsqueeze(0)*expand_sims).sum(dim=-1)
    dyds = (dypoly.unsqueeze(0)*expand_sims).sum(dim=-1)
    ddyds = (ddypoly.unsqueeze(0)*expand_sims).sum(dim=-1)

    psi = torch.atan2(dyds,dxds)

    den = dyds ** 2 + dxds ** 2
    psidot = (1./den) * (-dxds * ddyds + dyds * ddxds)


    simstates[...,0] = x
    simstates[...,1] = y
    simstates[...,2] = sdot
    simstates[...,3] = psi
    simstates[...,4] = psidot

    return simstates.reshape(T, -1)




def SVT_to_simstates(svt):
    """
    Converts a StackedVehicleTrajectory into Simulator states
    Args:
        svt (StackedVehicleTrajectory): stacked vehicle trajectory
    Returns:
        simstates (torch.tensor): (svt.Tind, svt.nv * 5) states
    """

    sims = torch.ones(svt.Tind, svt.nv) * np.nan
    simv = torch.ones(svt.Tind, svt.nv) * np.nan
    simstates = torch.ones(svt.Tind, svt.nv, 5) * np.nan

    for i in range(svt.nv):

        si = svt.s[i]
        vi = svt.v[i]
        ti = int(svt.t0[i]/svt.dt) - svt.minTind

        sims[ti:ti+len(si), i] = si
        simv[ti:ti+len(vi), i] = vi

    simstates = ssdot_to_simstates(sims, simv, 
                                svt.xpoly, svt.dxpoly, svt.ddxpoly,
                                svt.ypoly, svt.dypoly, svt.ddypoly)

    return simstates.reshape(svt.Tind, -1)










